@@ -26,8 +26,10 @@ from sklearn.ensemble import IsolationForest
 import pickle
 
 import warnings
+
 warnings.filterwarnings('ignore')
 
+
 def plot_feature_importance(importance, names, model_type, xlabel='SHAP values', title=' '):
     # Create arrays from feature importance and feature names
     feature_importance = np.array(importance)
@@ -1347,7 +1349,7 @@ def return_xy_cnt(x, y_target, S_bar_set, x_train, y_train, w):
         y_train_cnt = np.array(y_target).reshape(1, -1)
     else:
         x_train_cnt = np.array(x_train_cnt)
-    #     x_train_cnt[:, S_bar_set] = x[S_bar_set]
+        #     x_train_cnt[:, S_bar_set] = x[S_bar_set]
         y_train_cnt = np.array(x_train_cnt.shape[0] * [y_target])
 
     return x_train_cnt, y_train_cnt
@@ -1859,3 +1861,328 @@ class RunExperiments:
                            self.counterfactuals_samples_sdp_local[idt][l], False, False)
                 print('  ')
             print(' ')
+
+
+def return_xy_cnt_reg(x, y_target, down, up, S_bar_set, x_train, y_train, w):
+    """
+    return the observations with y=y_target that fall in the projected leaf of x
+    when we condition given S=S_bar of x.
+    """
+
+    x_train_cnt = []
+    y_train_cnt = []
+    for i, wi in enumerate(w):
+        if wi != 0 and down <= y_train[i] <= up:
+            x_train_cnt.append(x_train[i].copy())
+            y_train_cnt.append(y_train[i].copy())
+
+    x_train_cnt = np.array(x_train_cnt)
+    y_train_cnt = np.array(y_train_cnt)
+
+    return x_train_cnt, y_train_cnt
+
+
+def return_leaf_cnt_reg(ac_explainer, S_star, x_train_cnt, y_train_cnt, down, up, x_train, y_train, pi):
+    """
+    return the original leaves of the observations with y=y_target that fall in the projected leaf
+    when we condition given S=S_bar of x.
+    """
+    size = x_train_cnt.shape[0]
+    sdp, rules = ac_explainer.compute_cdp_rule(x_train_cnt, y_train_cnt, np.array(size * [down]), np.array(size * [up]),
+                                               x_train, y_train,
+                                               size * [list(range(x_train.shape[1]))]
+                                               )
+    #     print(sdp)
+    if np.sum(sdp >= pi) != 0:
+        rules_unique = np.unique(rules[sdp >= pi], axis=0)
+    else:
+        rules_unique = np.expand_dims(rules[np.argmax(sdp)], axis=0)
+
+    r_buf = rules_unique.copy()
+    for i in range(rules_unique.shape[0]):
+        list_ric = [r.copy() for r in r_buf if not np.allclose(r, rules_unique[i])]
+        find_union(rules_unique[i], list_ric, S=S_star)
+
+    return rules_unique
+
+
+def remove_in(ra):
+    """
+    remove A if A subset of B in the list of compatible leaves
+    """
+    for i in range(ra.shape[0]):
+        for j in range(ra.shape[0]):
+            if i != j and np.prod([(ra[i, s, 1] <= ra[j, s, 1]) * (ra[i, s, 0] >= ra[j, s, 0])
+                                   for s in range(ra.shape[1])], axis=0).astype(bool):
+                ra[i] = ra[j]
+    return np.unique(ra, axis=0)
+
+
+def get_compatible_leaf_reg(acvtree, x, y_target, down, up, S_star, S_bar_set, w, x_train, y_train, pi, acc_level):
+    """
+    Compute the compatible leaves and order given their accuracy
+    """
+    x_train_cnt, y_train_cnt = return_xy_cnt_reg(x, y_target, down, up, S_bar_set, x_train, y_train, w)
+    compatible_leaves = return_leaf_cnt_reg(acvtree, S_star, x_train_cnt, y_train_cnt, down, up, x_train, y_train, pi)
+    compatible_leaves = np.unique(compatible_leaves, axis=0)
+    compatible_leaves = remove_in(compatible_leaves)
+    compatible_leaves = np.round(compatible_leaves, 2)
+
+    partition_leaf = compatible_leaves.copy()
+    d = partition_leaf.shape[1]
+    nb_leaf = partition_leaf.shape[0]
+    leaves_acc = []
+    suf_leaf = []
+
+    for i in range(nb_leaf):
+        x_in = np.prod([(x_train[:, s] <= partition_leaf[i, s, 1]) * (x_train[:, s] > partition_leaf[i, s, 0])
+                        for s in range(d)], axis=0).astype(bool)
+
+        y_in = y_train[x_in]
+        acc = np.mean((down <= y_in) * (y_in <= up))
+        leaves_acc.append(acc)
+
+        if acc >= acc_level:
+            suf_leaf.append(partition_leaf[i])
+
+    best_id = np.argmax(leaves_acc)
+    return suf_leaf, partition_leaf, leaves_acc, partition_leaf[best_id], leaves_acc[best_id]
+
+
+def return_counterfactuals_reg(ac_explainer, suf_leaf, S_star, S_bar_set, x, y, down, up, x_train, y_train, pi_level):
+    """
+    Compute the SDP of each C_S and return the ones that has sdp >= pi_level
+    """
+    counterfactuals = []
+    counterfactuals_sdp = []
+    counterfactuals_w = []
+
+    for leaf in suf_leaf:
+
+        cond = np.ones(shape=(1, x_train.shape[1], 2))
+        cond[:, :, 0] = -1e+10
+        cond[:, :, 1] = 1e+10
+
+        for s in S_bar_set:
+            cond[:, s, 0] = x[:, s]
+            cond[:, s, 1] = x[:, s]
+
+        cond[:, S_star] = leaf[S_star]
+
+        size = x.shape[0]
+        sdp, w = ac_explainer.compute_cdp_cond_weights(x, y, np.array(size * [down]), np.array(size * [up]), x_train,
+                                                       y_train, S=[S_bar_set], cond=cond,
+                                                       pi_level=pi_level)
+        #         print(sdp)
+        if sdp >= pi_level:
+            counterfactuals.append(cond)
+            counterfactuals_sdp.append(sdp)
+            counterfactuals_w.append(w)
+
+    return np.unique(counterfactuals, axis=0), np.unique(counterfactuals_sdp, axis=0), \
+           np.unique(counterfactuals_w, axis=0)
+
+
+def return_global_counterfactuals_reg(ac_explainer, data, y_data, down, up, s_star, n_star, x_train, y_train, w,
+                                      acc_level, pi_level):
+    """
+    stack all to compute the C_S for each observations
+    """
+    N = data.shape[0]
+    suf_leaves = []
+    counterfactuals_samples = []
+    counterfactuals_samples_sdp = []
+    counterfactuals_samples_w = []
+
+    for i in tqdm(range(N)):
+        suf_leaf, _, _, _, _ = get_compatible_leaf_reg(ac_explainer, data[i], 1 - y_data[i], down[i], up[i], s_star[i],
+                                                       n_star[i], w[i],
+                                                       x_train, y_train, pi=pi_level, acc_level=acc_level)
+        suf_leaves.append(suf_leaf)
+        #         print(suf_leaf)
+        #         print(np.unique(suf_leaf, axis=0).shape, len(suf_leaf))
+        counterfactuals, counterfactuals_sdp, w_cond = \
+            return_counterfactuals_reg(ac_explainer, suf_leaf, s_star[i], n_star[i], data[i].reshape(1, -1),
+                                       y_data[i].reshape(1, -1), down[i], up[i], x_train, y_train, pi_level)
+
+        counterfactuals_samples.append(counterfactuals)
+        counterfactuals_samples_sdp.append(counterfactuals_sdp)
+        counterfactuals_samples_w.append(w_cond)
+    #     print(counterfactuals_samples)
+    return counterfactuals_samples, counterfactuals_samples_sdp, counterfactuals_samples_w
+
+
+def return_possible_values(rule, rule_data, sdp_all, pi_level):
+    d = rule.shape[0]
+    adj_rules = np.array([rule_data[i] for i in range(sdp_all.shape[0]) if sdp_all[i] >= pi_level])
+
+    left_by_var = []
+    right_by_var = []
+
+    for var in range(d):
+        left = set([adj_rules[i, var, 0] for i in range(adj_rules.shape[0]) if adj_rules[i, var, 0] <= rule[var, 0]])
+        right = set([adj_rules[i, var, 1] for i in range(adj_rules.shape[0]) if adj_rules[i, var, 1] >= rule[var, 1]])
+
+        left.add(rule[var, 0])
+        right.add(rule[var, 1])
+
+        left_by_var.append(list(left))
+        right_by_var.append(list(right))
+    return left_by_var, right_by_var
+
+
+def rdm_sample_fromlist(list_values):
+    return list_values[np.random.randint(0, len(list_values))]
+
+
+def generate_newrule(rule, left_by_var, right_by_var, p_best=0.6):
+    d = rule.shape[0]
+    rule_c = rule.copy()
+    if np.random.rand() > p_best:
+        for i in range(d):
+            rule_c[i, 0] = rdm_sample_fromlist(left_by_var[i])
+            rule_c[i, 1] = rdm_sample_fromlist(right_by_var[i])
+    else:
+        for i in range(d):
+            left = [left_by_var[i][j] for j in range(len(left_by_var[i])) if left_by_var[i][j] <= rule_c[i, 0]]
+            right = [right_by_var[i][j] for j in range(len(right_by_var[i])) if right_by_var[i][j] >= rule_c[i, 1]]
+
+            rule_c[i, 0] = rdm_sample_fromlist(left)
+            rule_c[i, 1] = rdm_sample_fromlist(right)
+
+    return rule_c
+
+
+def check_rule(rule, x_train, bad_sample):
+    x_in = np.prod([(x_train[:, s] <= rule[s, 1]) * (x_train[:, s] >= rule[s, 0])
+                    for s in range(x_train.shape[1])], axis=0)
+    return np.sum(x_in * bad_sample) == 0
+
+
+def generate_valid_rule(rule, left_by_var, right_by_var, x_train, bad_sample, p_best=0.9, n_try=100):
+    valid = False
+    it = 0
+    while valid == False and it <= n_try:
+        gen_rule = generate_newrule(rule, left_by_var, right_by_var, p_best)
+        valid = check_rule(gen_rule, x_train, bad_sample)
+        it += 1
+    if valid:
+        return gen_rule
+    else:
+        return rule
+
+
+def generate_rules(rule, left_by_var, right_by_var, x_train, bad_sample, p_best=0.6, n_try=100, n_gen=10):
+    rules = [generate_valid_rule(rule, left_by_var, right_by_var, x_train, bad_sample, p_best, n_try) for i in
+             range(n_gen)]
+    return rules
+
+
+def convert_boundvalues(rules, max_values, min_values):
+    for i in range(rules.shape[0]):
+        if rules[i, 0] <= -1e+10:
+            rules[i, 0] = min_values[i]
+        if rules[i, 1] <= -1e+10:
+            rules[i, 1] = min_values[i]
+        if rules[i, 0] >= 1e+10:
+            rules[i, 0] = max_values[i]
+        if rules[i, 1] >= 1e+10:
+            rules[i, 1] = max_values[i]
+    return rules
+
+
+def volume_rectangle(rec, max_values, min_values):
+    d = rec.shape[0]
+    rec_c = rec.copy()
+    rec_c = convert_boundvalues(rec_c, max_values, min_values)
+    v = 1
+    for i in range(d):
+        v *= (rec_c[i, 1] - rec_c[i, 0])
+    return v
+
+
+def volume_rectangles(recs, max_values, min_values):
+    return [volume_rectangle(recs[i], max_values, min_values) for i in range(len(recs))]
+
+
+def max_dens(recs, x_train, bad_sample):
+    d = x_train.shape[1]
+    x_in = np.prod([(x_train[:, s] <= recs[s, 1]) * (x_train[:, s] > recs[s, 0])
+                    for s in range(d)], axis=0)
+    return np.sum(x_in * (1 - bad_sample)) / np.sum(1 - bad_sample)
+
+
+def max_denss(recs, x_train, bad_sample):
+    return [max_dens(recs[i], x_train, bad_sample) for i in range(len(recs))]
+
+
+def rules_simulated_annealing(value_function, rule, left_by_var, right_by_var, x_train, bad_sample, p_best=0.9,
+                              n_try=100,
+                              batch=100, max_iter=200, temp=1,
+                              max_iter_convergence=50):
+    max_values = [np.max(x_train[:, i]) for i in range(x_train.shape[1])]
+    min_values = [np.min(x_train[:, i]) for i in range(x_train.shape[1])]
+
+    best = generate_rules(rule, left_by_var, right_by_var, x_train, bad_sample, p_best, n_try, n_gen=1)
+    best_eval = value_function(best, max_values, min_values)[0]
+    curr, curr_eval = np.squeeze(best[0]), best_eval
+
+    move = 0
+    for i in range(max_iter):
+
+        x_cand = generate_rules(curr, left_by_var, right_by_var, x_train, bad_sample, p_best, n_try, n_gen=batch)
+        score_candidates = value_function(x_cand, max_values, min_values)
+
+        candidate_eval = np.max(score_candidates)
+        candidate = x_cand[np.argmax(score_candidates)]
+
+        if candidate_eval > best_eval:
+            best, best_eval = candidate, candidate_eval
+            move = 0
+        else:
+            move += 1
+
+        # check convergence
+        if move > max_iter_convergence:
+            break
+
+        diff = candidate_eval - curr_eval
+        t = temp / np.log(float(i + 1))
+        metropolis = np.exp(-diff / t)
+
+        if diff > 0 or rand() < metropolis:
+            curr, curr_eval = candidate, candidate_eval
+
+    return best, best_eval
+
+
+def rules_by_annealing(value_function, rules, rules_data, sdp_all, x_train, pi_level=0.9, p_best=0.6, n_try=50,
+                       batch=100, max_iter=200, temp=1, max_iter_convergence=50):
+    sr, sr_eval = [], []
+    for idx in tqdm(range(rules.shape[0])):
+        rule = rules[idx]
+        rule_data = rules_data[idx]
+        bad_sample = 1. - (sdp_all[idx] >= pi_level)
+        sdp_rule = sdp_all[idx]
+
+        left_by_var, right_by_var = return_possible_values(rule, rule_data, sdp_rule, pi_level)
+        best, best_eval = rules_simulated_annealing(value_function, rule, left_by_var, right_by_var, x_train,
+                                                    bad_sample,
+                                                    p_best, n_try,
+                                                    batch, max_iter, temp, max_iter_convergence)
+        sr.append(np.squeeze(best))
+        sr_eval.append(best_eval)
+    return np.array(sr), sr_eval
+
+
+def remove_in_wsdp(ra, sa):
+    """
+    remove A if A subset of B in the list of compatible leaves
+    """
+    for i in range(ra.shape[0]):
+        for j in range(ra.shape[0]):
+            if i != j and np.prod([(ra[i, s, 1] <= ra[j, s, 1]) * (ra[i, s, 0] >= ra[j, s, 0])
+                                   for s in range(ra.shape[1])], axis=0).astype(bool):
+                ra[i] = ra[j]
+                sa[i] = sa[j]
+    return np.unique(ra, axis=0), np.unique(sa, axis=0)
